// ======================================================================
//
// Xform.h
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#ifndef XFORM_H
#define XFORM_H

// ======================================================================

#include "Vector.h"
class Quaternion;

// ======================================================================

// This class provides a 3x4 matrix for 3d positional and rotational transformations.  The 4th row is always assumed 
// to be [0, 0, 0, 1].  This class does not support scaling or shearing as part of the contents of the matrix.
// Requiring the matrices to be purely rotational and translational lets us invert the matrix by simply using
// the transpose of the upper 3x3 block of the matrix.
//
// All transforms are left-handed.  Positive rotations around an axis rotate in a clockwise fashion when viewed
// from the positive side of the axis being rotated about.
//
// Applying a series of rotations will introduce numerical error that causes the three vectors in the upper 3x3
// of the transform to become non-orthogonal and non-unit vectors.  Calling Xform::reorthogonalize() for a transform
// will reduce the numerical error by reconstructing the transform based on its current data.

// While it would be useful to explain the matrix and its contents here, it is currently not worth the time to write
// such documentation.  Please refer to Foley & van Dam, or some such other useful reference.

class Xform
{
	friend class Quaternion;
	friend class Geometry;

public:

	// The multiplicative identity matrix, consisting of 0's, with a diagonal of 1's.
	static const Xform identity;

public:

	enum
	{
		SAVE_SIZE = 12
	};

	struct Save
	{
		real save[SAVE_SIZE];
	};

private:

	real matrix[3][4];

private:
	
	void   transpose3x3(const Xform &m);
	void   makeScale(const Vector &scale);

public:
	
	// default copy constructor and assignment operators are valid, so let the compiler generate them

	Xform(void);

	// specific matrix creation routines
	void   identify(void);

	void   invert(const Xform &m);
	void   reorthogonalize(void);

	void   getPosition(Vector *v) const;
	void   getOrientationYPR(real *yawRadians, real *pitchRadians, real *rollRadians) const;

	// frame vector routines
	void   frameI(Vector *i) const;
	void   frameJ(Vector *j) const;
	void   frameK(Vector *k) const;
	void   inverseFrameI(Vector *i) const;
	void   inverseFrameJ(Vector *j) const;
	void   inverseFrameK(Vector *k) const;

	// translational changes
	void   moveAbsolute(real x, real y, real z);
	void   moveAbsolute(const Vector &v);
	void   moveRelative(real x, real y, real z);
	void   moveRelative(const Vector &v);
	void   move(real x, real y, real z);
	void   move(const Vector &v);
	
	// rotational changes
	void   rotateReset(void);
	void   pitch(real radians);
	void   yaw(real radians);
	void   roll(real radians);
	void   rotateToFrame(const Vector *unitI, const Vector *unitJ, const Vector *unitK, bool needsReorthogonalize=true);

	// math routines
	void   multiply(const Xform &src1, const Xform &src2);

	// vector transformation routines
	void   rotate(const Vector &src, Vector *dest) const;
	void   rotateTranslate(const Vector &src, Vector *dest) const;
	void   translate(const Vector &src, Vector *dest) const;
	void   inverseRotate(const Vector &src, Vector *dest) const;
	void   inverseRotateTranslate(const Vector &src, Vector *dest) const;
	void   rotateTranslate(const Vector *src, Vector *dest, int cnt) const;
	void   rotateTranslatePlane(const Vector &srcNormal, real srcPlaneD, Vector *destNormal, real *destPlaneD) const;

	// specialized vector transformation routines for rendering
	bool   rotateTranslatePerspectiveClip(const Vector *src, Vector *dest, int cnt, real nearPlane, real farPlane, real xMult, real yMult, int8 *clip) const;

	// support routines
	void   writeSave(Save *save);
	void   readSave(const Save &save);
};

// ======================================================================
// Make this the multiplicative identity matrix
//
// Remarks:
//
//   This matrix sets the elements on the diagonal to 1's and all other elements to 0's.

inline void Xform::identify(void)
{
	matrix[0][0] = CONST_REAL(1);
	matrix[0][1] = CONST_REAL(0);
	matrix[0][2] = CONST_REAL(0);
	matrix[0][3] = CONST_REAL(0);

	matrix[1][0] = CONST_REAL(0);
	matrix[1][1] = CONST_REAL(1);
	matrix[1][2] = CONST_REAL(0);
	matrix[1][3] = CONST_REAL(0);

	matrix[2][0] = CONST_REAL(0);
	matrix[2][1] = CONST_REAL(0);
	matrix[2][2] = CONST_REAL(1);
	matrix[2][3] = CONST_REAL(0);
}

// ----------------------------------------------------------------------
// Construct a transform and set it to the multiplicative identity transform
//
// Remarks:
//
//   Since transform matrices are most commonly multiplied together, the multiplicative identity is the most
//   reasonable default value for a transform.

inline Xform::Xform(void)
{
	identify();
}

// ----------------------------------------------------------------------
// Return the position in space of this matrix
//
// Remarks:
//
//    This routine takes a pointer to the vector to receive the position instead of returning a temporary vector
//    for efficiency reasons.

inline void Xform::getPosition(
	Vector *v  // the vector to receive the position
	) const
{
	v->x = matrix[0][3];
	v->y = matrix[1][3];
	v->z = matrix[2][3];
}

// ----------------------------------------------------------------------
// Reset this transform to have no rotation
//
// Remarks:
//
//   This routine is much like identify, but does not reset the position.  It sets the upper 3x3 matrix to the
//   multiplicative identity.

inline void Xform::rotateReset(void)
{
	matrix[0][0] = CONST_REAL(1);
	matrix[0][1] = CONST_REAL(0);
	matrix[0][2] = CONST_REAL(0);

	matrix[1][0] = CONST_REAL(0);
	matrix[1][1] = CONST_REAL(1);
	matrix[1][2] = CONST_REAL(0);

	matrix[2][0] = CONST_REAL(0);
	matrix[2][1] = CONST_REAL(0);
	matrix[2][2] = CONST_REAL(1);
}

// ----------------------------------------------------------------------
// Set the position of this transform
//
// See also:
//
//   moveRelative(), move()

inline void Xform::moveAbsolute(
	real px,  // new X position
	real py,  // new Y position
	real pz   // new Z position
	)
{
	matrix[0][3] = px;
	matrix[1][3] = py;
	matrix[2][3] = pz;
}

// ----------------------------------------------------------------------
// Set the absolute position of this transform
//
// See also:
//
//   moveRelative(), move()

inline void Xform::moveAbsolute(
	const Vector &v  // new position
	)
{
	matrix[0][3] = v.x;
	matrix[1][3] = v.y;
	matrix[2][3] = v.z;
}

// ----------------------------------------------------------------------
// Adds the specified offset to the position of this transform
//
// See also:
//
//   moveAbsolute(), move()

inline void Xform::moveRelative(
	real dx,  // delta to move in the X direction
	real dy,  // delta to move in the Y direction
	real dz   // delta to move in the Z direction
	)
{
	matrix[0][3] += dx;
	matrix[1][3] += dy;
	matrix[2][3] += dz;
}

// ----------------------------------------------------------------------
// Adds the specified offset to the position of this transform
//
// See also:
//
//   moveAbsolute(), move()

inline void Xform::moveRelative(
	const Vector &v  // delta to move
	)
{
	matrix[0][3] += v.x;
	matrix[1][3] += v.y;
	matrix[2][3] += v.z;
}

// ------------------------------------------------------------------------
// Retrieve the vector that points along the X axis of this transform's frame
//
// Remarks:
//
//   The result will be a unit vector (providing the transform has not had too many rotations applied without 
//   reorthogonalize() called).
// 
// See Also:
//
//   reorthogonalize()

inline void Xform::frameI(
	Vector *i  // vector to receive the X axis
	) const
{
	i->x = matrix[0][0];
	i->y = matrix[1][0];
	i->z = matrix[2][0];
}

// ------------------------------------------------------------------------
// Retrieve the vector that points along the Y axis of this transform's frame
//
// Remarks:
//
//   The result will be a unit vector (providing the transform has not had too many rotations applied without 
//   reorthogonalize() called).
// 
// See Also:
//
//   reorthogonalize()

inline void Xform::frameJ(
	Vector *j // vector to receive the Y axis
	) const
{
	j->x = matrix[0][1];
	j->y = matrix[1][1];
	j->z = matrix[2][1];
}

// ------------------------------------------------------------------------
// Retrieve the vector that points along the Z axis of this transform's frame
//
// Remarks:
//
//   The result will be a unit vector (providing the transform has not had too many rotations applied without 
//   reorthogonalize() called).
// 
// See Also:
//
//   reorthogonalize()

inline void Xform::frameK(
	Vector *k  // vector to receive the Z axis
	) const
{
	k->x = matrix[0][2];
	k->y = matrix[1][2];
	k->z = matrix[2][2];
}

// ------------------------------------------------------------------------
// Retrieve the vector that points along the X axis of the parent space
//
// Remarks:
//
//   The result will be a unit vector (providing the transform has not had too many rotations applied without 
//   reorthogonalize() called).
// 
// See Also:
//
//   reorthogonalize()

inline void Xform::inverseFrameI(Vector *i) const
{
	i->x = matrix[0][0];
	i->y = matrix[0][1];
	i->z = matrix[0][2];
}

// ------------------------------------------------------------------------
// Retrieve the vector that points along the Y axis of the parent space
//
// Remarks:
//
//   The result will be a unit vector (providing the transform has not had too many rotations applied without 
//   reorthogonalize() called).
// 
// See Also:
//
//   reorthogonalize()

inline void Xform::inverseFrameJ(Vector *j) const
{
	j->x = matrix[1][0];
	j->y = matrix[1][1];
	j->z = matrix[1][2];
}

// ------------------------------------------------------------------------
// Retrieve the vector that points along the Z axis of the parent space
//
// Remarks:
//
//   The result will be a unit vector (providing the transform has not had too many rotations applied without 
//   reorthogonalize() called).
// 
// See Also:
//
//   reorthogonalize()

inline void Xform::inverseFrameK(Vector *k) const
{
	k->x = matrix[2][0];
	k->y = matrix[2][1];
	k->z = matrix[2][2];
}

// ======================================================================

#endif
